@@ -1,10 +1,10 @@
 from . import (
+    _utils,
     certronic,
     exactian,
     nettime5,
     nettime6,
     specmanagerapi,
     specmanagerdb,
-    visma,
-    utils
+    visma
 )
\ No newline at end of file
