@@ -1,36 +1,31 @@
-from string import digits as str_digits, ascii_lowercase as str_letters
-from random import choice as r_choice
-from urllib.parse import urlparse, urlencode, urljoin
-from base64 import b64encode, b64decode
+from __future__ import annotations
+from typing import List, Optional
+
+import datetime as dt
 import requests
-import datetime
-import re
+from urllib.parse import urlparse, urljoin
+from base64 import b64encode, b64decode
 
-__nettime__ = "6.0.1.17769"
+try:
+    from ._utils import Decorators, create_random_suffix
+except ImportError:
+    from _utils import Decorators, create_random_suffix
 
-def random_str(size=5, chars=str_digits + str_letters):
-    """ Return a str of 'size' len with numbers and ascii lower letters. """
+__nettime__ = "6.0.1.19025"
 
-    return ''.join(r_choice(chars) for _ in range(size))
-
-def create_random_suffix(name=""):
-    """ Create a random name adding suffix after of clean recived name. """
-
-    clean = re.sub('[^a-zA-Z0-9]', '_', name)
-    clean += '_' if name else "" 
-    clean += random_str(size=5)
-
-    return clean
 
 class Query:
 
-    def __init__(self, fields: list, \
-            startDate: str = datetime.date.today().isoformat(), \
-            filterExp: str = ""):
+    def __init__(
+        self,
+        fields: list,
+        startDate: Optional[str] = dt.date.today().isoformat(),
+        filterExp: Optional[str] = ""
+    ) -> None:
         self.queryfields = self.QueryFields(fields, startDate)
         self.filterExp = self.filter_prepare(expression=filterExp)
 
-    def prepare(self):
+    def prepare(self) -> str:
         """ Format a query in str for use in url. """
 
         query = '{}"fields":{}'.format('{', self.queryfields.prepare())
@@ -41,16 +36,19 @@ class Query:
         query += '}'
         return query
 
-    def filter_prepare(self, expression: str = ""):
+    def filter_prepare(self, expression: Optional[str] = "") -> str:
         return expression.replace('"', "'")
 
     class QueryFields:
-        def __init__(self, names: list, \
-                startDate: str = datetime.date.today().isoformat()):
+        def __init__(
+            self,
+            names: List[str],
+            startDate: Optional[str] = dt.date.today().isoformat()
+        ) -> None:
             self.names = names
             self.startDate = startDate
 
-        def prepare(self):
+        def prepare(self) -> str:
             """ Format a query in str for use in url. """
 
             fields = []
@@ -65,30 +63,42 @@ class Query:
 
 class Client:
 
-    def __init__(self, url: str, username: str, pwd: str, *args, **kwargs):
-        """
-        Create a conection with nettime app using recived parameters.
+    def __init__(
+        self,
+        url: str,
+        username: str,
+        pwd: str,
+        session: Optional[requests.Session] = None,
+        **kwargs
+    ) -> None:
+        """ Create a connector with nettime app using recived parameters
 
-        :param url: Nettime url. Eg https://server-name:8091/.
-        :param username: Nettime username.
-        :param pwd: Nettime password.
-
-        :return: :class:`Client` object
-        :rtype: Client
+        Args:
+            url (str): Nettime url. Eg "https://server-name:8091/"
+            username (str): Nettime username
+            pwd (str): Nettime password
         """
 
-        super().__init__(*args, **kwargs)
         self.client_url = urlparse(url)
         self.username = username
         self.pwd = b64encode(pwd.encode('utf-8'))
+        
+        # base headers
+        self.headers = {
+            "DNT": "1",
+            # "Content-Type": "application/json;charset=UTF-8",
+            "Accept-Encoding": "gzip,deflate",
+            # "Cookie": f"sessionID={self.access_token}; i18next=es",
+        }
+        
+        # manual session or None
+        self.session = session
 
         ### None values
         self.access_token = None
-        self.headers = None
         self.user_rol = None
+        self.settings = None
 
-        #connect client automatically
-        self.connect()
 
     def __str__(self):
         return '{}{} en {}'.format(
@@ -105,55 +115,69 @@ class Client:
             b64decode(self.pwd).decode('utf-8'),
         )
 
-    def __enter__(self, *args, **kwargs):
+    def __enter__(self, *args, **kwargs) -> Client:
+        self.start_session()
         return self
 
-    def __exit__(self, *args, **kwargs):
-        return self.disconnect()
+    def __exit__(self, *args, **kwargs) -> None:
+        self.close_session()
+
+    def refresh_headers(
+        self,
+        access_token: Optional[str] = None,
+        remove_token: Optional[bool] = False,
+        remove_content_type: Optional[bool] = False
+    ) -> None:
+        if remove_token:
+            self.access_token = None
+            self.headers.pop("Cookie", None)
+            self.session.headers.pop("Cookie", None)
+
+        if remove_content_type:
+            self.headers.pop("Content-Type", None)
+            self.session.headers.pop("Content-Type", None)
+        
+        if access_token:
+            self.access_token = access_token
+
+        if self.access_token:
+            self.headers.update({
+                "Cookie": f"sessionID={self.access_token}; i18next=es"
+            })
+
+        if "Content-Type" not in self.headers and not remove_content_type:
+            self.headers.update({
+                "Content-Type": "application/json;charset=UTF-8"
+            })
+
+    def start_session(self) -> None:
+        self.session = requests.Session()
+        self.session.headers.update(self.headers)
+
+        # login
+        self.login()
+
+    def refresh_session(self) -> None:
+        self.session.headers.update(self.headers)
+
+    def close_session(self):
+        self.logout()
+        self.session.close()
+        self.session = None
 
     @property
     def is_connected(self):
         """ Informs if client has headers and access_token. """
 
-        return bool(self.headers) and bool(self.access_token)
+        return bool("Cookie" in self.headers and bool(self.access_token))
 
-    def connect(self):
-        """ Connect the client to set access_token and headers values. """
-
-        if self.is_connected:
-            return
-
-        # url and data prepare
-        url = urljoin(self.client_url.geturl(), '/api/login')
-        data = {
-            "username": self.username,
-            "pwd": b64decode(self.pwd).decode('utf-8'),
-        }
-        response = requests.post(url, data=data)
-
-        if response.status_code not in range(200, 300):
-            raise ConnectionError(response.text)
-
-        json_data = response.json()
-
-        if not json_data.get("ok"):
-            raise ConnectionError(json_data.get("message"))
-
-        self.access_token = json_data.get("access_token")
-        self.headers = self.get_headers()
-        self.settings = self.get_settings()
-        self.user_rol = self.get_user_rol()
-
-    def reconnect(self):
-        """ Reconnect client cleaning headers and access_token. """
-
-        #clean token and headers for safety
-        self.access_token = None
-        self.headers = None
-
-        self.connect()
-
-    def get(self, path: str, params: dict = None, **kwargs):
+    @Decorators.ensure_session
+    def get(
+        self,
+        path: str,
+        params: dict = None,
+        **kwargs
+    ) -> Union[Dict, List, requests.Response]:
         """
         Sends a GET request to nettime url.
 
@@ -165,29 +189,23 @@ class Client:
         :rtype: json
         """
 
-        if not self.is_connected:
-            raise ConnectionError("Cliente desconectado. Utilice connect().")
-        
-        # query prepare
-        query = {
-            "url": urljoin(self.client_url.geturl(), path),
-            "params": params,
-            "headers": self.headers,
-            "timeout": kwargs.get("timeout", 10),
-            "stream": kwargs.get("stream", False)
-        }
+        # prepare url
+        url = urljoin(self.client_url.geturl(), path)
 
         # consulting nettime
-        response = requests.get(**query)
+        response = self.session.get(url=url, params=params, **kwargs)
 
         # if session was closed, reconect client and try again
         if response.status_code == 401:
-            self.reconnect()
-            return self.get(path, params=params, **kwargs)
+            self.relogin()
+            return self.get(path=path, params=params, **kwargs)
 
         # raise if was an error
         if response.status_code not in range(200, 300):
-            raise ConnectionError(response.text)
+            raise ConnectionError({
+                "status": response.status_code,
+                "detail": response.text
+            })
         
         # if request is stream type, return all response
         if kwargs.get("stream"):
@@ -196,7 +214,15 @@ class Client:
         # to json
         return response.json()
 
-    def post(self, path, data=None, json=None, **kwargs):
+    @Decorators.ensure_session
+    def post(
+        self,
+        path: str,
+        params: dict = None,
+        data: dict = None,
+        json: dict = None,
+        **kwargs
+    ) -> Union[Dict, List]:
         """
         Sends a POST request to nettime url.
 
@@ -210,30 +236,35 @@ class Client:
         :rtype: json
         """
 
-        # wait active conection
-        if not self.is_connected:
-            raise ConnectionError("Cliente desconectado. Utilice connect().")
-
-        # query prepare
-        query = {
-            "url": urljoin(self.client_url.geturl(), path),
-            "data": data,
-            "json": json,
-            "headers": self.headers,
-            "timeout": kwargs.get("timeout", 10)
-        }
+        # prepare url
+        url = urljoin(self.client_url.geturl(), path)
 
         # consulting nettime
-        response = requests.post(**query)
+        response = self.session.post(
+            url=url,
+            params=params,
+            data=data,
+            json=json,
+            **kwargs
+        )
 
         # if session was closed, reconect client and try again
         if response.status_code == 401:
-            self.reconnect()
-            return self.post(path, data=data, json=json, **kwargs)
+            self.relogin()
+            return self.post(
+                path=path,
+                params=params,
+                data=data,
+                json=json,
+                **kwargs
+            )
 
         # raise if was an error
         if response.status_code not in range(200, 300):
-            raise ConnectionError(response.status_code, response.text)
+            raise ConnectionError({
+                "status": response.status_code,
+                "detail": response.text
+            })
 
         # to json -> json
         json_response = response.json()
@@ -246,31 +277,58 @@ class Client:
         # return json response
         return json_response
 
-    def get_headers(self):
-        """ Return headers for a specific conection """
+    def login(self) -> None:
+        """ Connect the client to set access_token and headers values. """
 
-        if not self.access_token:
-            raise ConnectionError("El cliente esta desconectado.")
+        if self.is_connected:
+            return
 
-        return {
-            "DNT": "1",
-            "Content-Type": "application/json;charset=UTF-8",
-            "Accept-Encoding": "gzip,deflate",
-            "Cookie": f"sessionID={self.access_token}; i18next=es",
+        # remove content type from headers. Must be not json
+        self.refresh_headers(remove_token=True, remove_content_type=True)
+        
+        # data prepare
+        data = {
+            "username": self.username,
+            "pwd": b64decode(self.pwd).decode('utf-8'),
         }
 
-    def disconnect(self):
+        # consulting nettime
+        json_data = self.post(path='/api/login', data=data)
+
+        if not json_data.get("ok", None):
+            raise ConnectionError({
+                "status": 401,
+                "detail": json_data.get("message")
+            })
+
+        # update access token and headers
+        self.refresh_headers(access_token=json_data.get("access_token"))
+        
+        # refresh session with updated headers
+        self.refresh_session()
+
+        self.settings = self.get_settings()
+        self.user_rol = self.get_user_rol()
+
+    def logout(self) -> None:
         """ Disconnect a client to clean the access_token. """
 
         if not self.is_connected:
             return
-        
+
         # disconnect ...
         response = self.post(path='/api/logout')
 
-        # reset values
-        self.access_token = None
-        self.headers = None
+        # clean token and headers for safety
+        self.refresh_headers(remove_token=True, remove_content_type=True)
+        self.refresh_session()
+
+    def relogin(self):
+        """ Reconnect client cleaning headers and access_token. """
+
+        # logout and login. Will fail if has no token
+        logout_result = self.logout()
+        login_result = self.login()
 
     def get_settings(self):
         """ Get settings of netTime Server. """
@@ -284,10 +342,10 @@ class Client:
 
     def get_days_offset(self, days: list):
         """ 
-        Convert a list of datetime.date or str format to int offset with 
+        Convert a list ofdt.date or str format to int offset with 
         self.setting.firstDate.
 
-        :param days: list of datetime.date or formatted str to get int offsets.
+        :param days: list ofdt.date or formatted str to get int offsets.
         
         :return: :class:`list` object
         :rtype: list
@@ -302,14 +360,14 @@ class Client:
             raise RuntimeError("No se puede obtener el setting firstDate.")
         
         # set first_date
-        first_date = datetime.date(firstYear, 1, 1)
+        first_date =dt.date(firstYear, 1, 1)
 
         # process dates
         days_numbers = []
         for day in days:
-            # ensure datetime.date type
-            if not isinstance(day, datetime.date):
-                day = datetime.date.fromisoformat(day)
+            # ensuredt.date type
+            if not isinstance(day,dt.date):
+                day =dt.date.fromisoformat(day)
 
             delta = day - first_date
             days_numbers.append(delta.days)
@@ -318,13 +376,13 @@ class Client:
 
     def get_days_from_offsets(self, offsets: list):
         """ 
-        Convert a list of int offset to datetime.date or str format with \
+        Convert a list of int offset todt.date or str format with \
         self.setting.firstDate.
 
         :param offsets: list of int to get dates.
         
-        :return: :class:`datetime.date` object
-        :rtype: datetime.date
+        :return: :class:dt.date` object
+        :rtype:dt.date
         """
 
         # wait active conection
@@ -336,12 +394,12 @@ class Client:
             raise RuntimeError("No se puede obtener el setting firstDate.")
 
         # set first_date
-        first_date = datetime.date(firstYear, 1, 1)
+        first_date =dt.date(firstYear, 1, 1)
 
         # process dates
         dates = []
         for offset in offsets:
-            date = first_date + datetime.timedelta(days=offset)
+            date = first_date +dt.timedelta(days=offset)
             dates.append(date.isoformat())
         
         return dates
@@ -416,7 +474,7 @@ class Client:
 
         return json_response
 
-    def get_employees(self, query=Query(["id", "nif"]), *args, **kwargs):
+    def get_employees(self, query=Query(["id", "nif"]), **kwargs):
         """
         Get employees from nettime. 
         
@@ -431,7 +489,12 @@ class Client:
 
         # use get general propose
         employees = self.get_elements(
-            container="Persona", query=query, *args, **kwargs)
+            container="Persona",
+            query=query,
+            **kwargs
+        )
+
+        # TODO: Check if recurssion is necessary
 
         return employees
 
@@ -553,8 +616,8 @@ class Client:
         return response[0]
 
     def get_day_info(self, employee: int, \
-            _from: str = datetime.date.today().isoformat(), \
-            to: str = datetime.date.today().isoformat()):
+            _from: str =dt.date.today().isoformat(), \
+            to: str =dt.date.today().isoformat()):
         """ 
         Get info, days, shifts, and results for a employe in a specific period. 
         """
@@ -570,8 +633,8 @@ class Client:
         return self.get(path='/api/day/results', params=params)
 
     def get_access_clockings(self, employee: int, \
-            _from: str = datetime.date.today().isoformat(), \
-            to: str = datetime.date.today().isoformat()):
+            _from: str =dt.date.today().isoformat(), \
+            to: str =dt.date.today().isoformat()):
         """ Get access clockings for a employe in a specific period. """
 
         # prepare task parameters
@@ -612,8 +675,8 @@ class Client:
         return self.get(path='/api/async/response', params=params)
 
     def get_results(self, employee: int, \
-            _from: str = datetime.date.today().isoformat(), \
-            to: str = datetime.date.today().isoformat()):
+            _from: str =dt.date.today().isoformat(), \
+            to: str =dt.date.today().isoformat()):
         """ Get results of day for a employee. """
 
         # prepare task parameters
@@ -629,13 +692,13 @@ class Client:
         # get task results
         return self.get_task_response(async_task.get('taskId'))
 
-    def clocking_prepare(self, employee: int, date_time: datetime.datetime, \
+    def clocking_prepare(self, employee: int, date_time:dt.datetime, \
             reader: int = -1, clocking_id: int = -1, action: str = None):
         """ Return a dict element with recived data. """
 
-        # ensure datetime.datetime type
-        if not isinstance(date_time, datetime.datetime):
-            date_time = datetime.datetime.fromisoformat(date_time)
+        # ensuredt.datetime type
+        if not isinstance(date_time,dt.datetime):
+            date_time =dt.datetime.fromisoformat(date_time)
 
         # prepare structure
         clocking_data = {
@@ -659,13 +722,13 @@ class Client:
             raise ValueError("Método no soportado para el tipo 'Persona'.")
 
         #teime process
-        if not isinstance(date, datetime.date):
-            date = datetime.date.fromisoformat(date)
+        if not isinstance(date,dt.date):
+            date =dt.date.fromisoformat(date)
 
-        if not isinstance(date, datetime.time):
-            time = datetime.datetime.strptime(time, "%H:%M").time()
+        if not isinstance(date,dt.time):
+            time =dt.datetime.strptime(time, "%H:%M").time()
 
-        date_time = datetime.datetime.combine(date, time)
+        date_time =dt.datetime.combine(date, time)
 
         json_data = {
             "idEmp": employee,
@@ -684,7 +747,7 @@ class Client:
         return self.post(path='/api/day/post/', json=json_data)
 
     def get_day_clockings(self, employee: int, \
-            date: str = datetime.date.today().isoformat()):
+            date: str =dt.date.today().isoformat()):
         """
         Get all the clockings (Horario) of an employe in a specific day.
         ** This method use portal API.
@@ -990,9 +1053,17 @@ class Client:
 
         return readers
 
-    def import_employee(self, structure: dict, **kwargs):
+    def import_employee(self, structure: dict, identifier: str = 'nif', \
+            **kwargs):
         """
         Create an employe from a structure. Update if exists, create if not.
+
+        :param structure: Dict with employee structure, like propertie: value.
+        :param identifier: (Optional) String with nettime field name to identiy
+            employee.
+        
+        :return: :class:`dict` object
+        :rtype: json
         """
 
         # employee structure
@@ -1000,14 +1071,14 @@ class Client:
 
         # search employee by nif
         query = Query(
-            fields=["id", "nif"],
-            filterExp=f'this.nif = "{structure.get("nif")}"',
+            fields=["id", identifier],
+            filterExp=f'this.{identifier} = "{structure.get(identifier)}"',
         )
         results = self.get_employees(query=query)
 
         # safety only
         if results.get('total') > 1:
-            raise ValueError("Más de un empleado con el mismo DNI.")
+            raise ValueError("Más de un empleado con el mismo identificador.")
 
         # update employee
         if results.get('total') == 1:
