@@ -1,54 +1,14 @@
-#from string import digits as str_digits, ascii_lowercase as str_letters
-#from random import choice as r_choice
-from urllib.parse import urlparse, urlencode, urljoin
+from concurrent.futures import ThreadPoolExecutor, as_completed
+from urllib.parse import urlparse, urljoin
 from base64 import b64encode, b64decode
 import requests
 import datetime
-import re
 import math
-from concurrent.futures import ThreadPoolExecutor, as_completed
 
 
-def fetch(client, method: str, params: dict = {}):
-    """ Execute method of a client with recived parameters """
-    method = getattr(client, method)
-    return method(**params)
+# TODO: Change to use requests sessions
 
 
-def bulk_get_pages(client, method, params: dict = {},
-        page_range: tuple = (1, 2), max_workers: int = 50, **kwargs) -> list:
-    """ Get a range of pages of a client simultaneously. """
-
-    # structures prepare
-    responses = []
-    processes = []
-    params_ = params
-
-    # create executor context
-    with ThreadPoolExecutor(max_workers=max_workers) as executor:
-
-        # processes prepare
-        for i in range(*page_range):
-            params_["params"]["page"] = i
-            processes.append(
-                executor.submit(
-                    fetch,
-                    client=client,
-                    method=method,
-                    params=params_
-                )
-            )
-
-        # execute and put result in out structure
-        [responses.append(r_.result()) for r_ in as_completed(processes)]
-
-        # close executor
-        executor.shutdown(wait=True)
-
-    # return responses
-    # list of pages result
-    return responses
-
 class JsonObject:
     """ Convert a dict['element'] for access like object.property. """
 
@@ -261,6 +221,60 @@ class Client:
         except:
             return {}
 
+    def bulk_get(self, path: str, params: dict = {}, \
+            together: tuple = tuple(), max_workers: int = 50, \
+            response_to_dict: bool = False, **kwargs):
+        """
+        Get from client object simultaneously.
+        
+        :param path: path to add to URL for the new :class:`Request` object.
+        :param params: (optional) Dictionary, list of tuples or bytes to send
+            in the query string for the :class:`Request`.
+        :param together: Iterator of dicts to get together. E.g. 
+            [{
+                'page': 1
+            },{
+                'page': 2
+            }]
+        :max_workers: int to specify max workers number to run simultaneously.
+        :response_to_dict: bool to group responses if there are dict objects.
+        :param \*\*kwargs: Optional arguments that ``request`` takes.
+        :return: :class:`dict` object
+        :rtype: dict
+        """
+
+        # structures prepare
+        responses = {} if response_to_dict else []
+        processes = []
+        params_ = params
+
+        # create executor context
+        with ThreadPoolExecutor(max_workers=max_workers) as executor:
+            # processes prepare
+            for e_ in together:
+                params_.update(e_)
+                processes.append(
+                    executor.submit(
+                        self.get,
+                        path=path,
+                        params=params_,
+                        **kwargs
+                    )
+                )
+
+            # execute and put result in out structure
+            if response_to_dict:
+                [responses.update(r.result()) for r in as_completed(processes)]
+            else:
+                [responses.append(r.result()) for r in as_completed(processes)]
+
+            # close executor
+            executor.shutdown(wait=True)
+
+        # return responses
+        # list of pages result
+        return responses
+
     def post(self, path, data=None, json=None, **kwargs):
         """
         Sends a POST request to visma url.
@@ -345,7 +359,7 @@ class Client:
         self.connect()
 
     def get_employees(self, employee: str = None, extension: str = None, \
-            all_pages: bool = False, **kwargs):
+            all_pages: bool = False, bulk_get: bool = False, **kwargs):
         """
         Use the endpoint to obtain the employees with the received data.
         
@@ -396,25 +410,36 @@ class Client:
             num_pages = math.ceil(
                 response.get('totalCount') / params.get('pageSize')
             )
+            
+            # TODO: fix it
+            # only with bulk_get param
+            if bulk_get:
+                together = [{'page': i} for i in range(2, num_pages + 1)]
 
-            # get list of pages
-            pages_result = bulk_get_pages(
-                client=self,
-                method='get',
-                params={
-                    'path': path,
-                    'params': params
-                },
-                page_range=(2, num_pages + 1),
-                max_workers=kwargs.get('max_workers', 50)
-            )
+                # get list of pages
+                pages_result = self.bulk_get(
+                    path=path,
+                    params=params,
+                    together=together,
+                    max_workers=kwargs.get('max_workers', 50)
+                )
+
+                # get values and add to structure
+                for pr in pages_result:
+                    response['values'].extend(pr.get('values'))
+
+            else:
+                # recursive get and extend response values
+                for i in range(2, num_pages + 1):
+                    # update page
+                    params['page'] = i
+                    response['values'].extend(
+                        self.get(path=path, params=params).get('values')
+                    )
 
-            # get values and add to structure
-            for pr in pages_result:
-                response['values'].extend(pr.get('values'))
         
         # put elements in dict if is ask extensions
-        if extension and kwargs.get('group_results', None):
+        if bulk_get and extension and kwargs.get('group_results', None):
             return {employee: {extension: response}}
 
         # return elements
